/*===========================================================================



Mini-DDS
========



Overview:
=========

CPU         :   PIC16F84A
Clock       :   HS crystal oscillator, 9.830400 MHz
Compiler    :   Hitech PICC
Version     :   1.00
Date        :   26.02.2009
Author      :   Iacopo Giangrandi



I/O connections:
================

PORTA        <-- FREQ_SEL       5-bit binary coded frequency selection (from predefined values).
PORTB        --> OUTPUT         8-bit wave output.



Description:
============

This implements a simple low frequency DDS generator. The frequency step is
10 Hz, the sampling frequency is 9600 Hz, the maximum output frequency is
4800 Hz and the resolution is 8 bits. Changing the binary values of PORTA
pins, the frequency can be selected from some predefined values.



============================================================================= */



// --------------------------------------------------------------------------
// External files inclusions.
// --------------------------------------------------------------------------
#include <pic.h>                                                // Uses a PIC16F84A




// --------------------------------------------------------------------------
// Configuration bits definitions.
// --------------------------------------------------------------------------
__CONFIG(HS & WDTDIS & PWRTEN & UNPROTECT);
// Use high speed crystal oscillator, do not use watchdog, enable power up
// timer and do not use data protection.




// --------------------------------------------------------------------------
// I/O definitions.
// --------------------------------------------------------------------------
#define OUTPUT      PORTB                                       // Output is sent directly on PORTB (8-bit)
#define FREQ_SEL    PORTA                                       // PORTA is used to binary select the frequency.




// --------------------------------------------------------------------------
// Global constants.
// --------------------------------------------------------------------------
const unsigned char cos_data[] =
      { 127, 127, 127, 127, 127, 127, 127, 127, 127, 127, 
        127, 127, 127, 127, 126, 126, 126, 126, 126, 126, 
        126, 126, 126, 126, 125, 125, 125, 125, 125, 125, 
        125, 124, 124, 124, 124, 124, 123, 123, 123, 123, 
        123, 122, 122, 122, 122, 122, 121, 121, 121, 121, 
        120, 120, 120, 119, 119, 119, 119, 118, 118, 118, 
        117, 117, 117, 116, 116, 116, 115, 115, 115, 114, 
        114, 114, 113, 113, 112, 112, 112, 111, 111, 110, 
        110, 110, 109, 109, 108, 108, 107, 107, 107, 106, 
        106, 105, 105, 104, 104, 103, 103, 102, 102, 101, 
        101, 100, 100,  99,  99,  98,  98,  97,  97,  96, 
         95,  95,  94,  94,  93,  93,  92,  92,  91,  90, 
         90,  89,  89,  88,  87,  87,  86,  86,  85,  84, 
         84,  83,  82,  82,  81,  81,  80,  79,  79,  78, 
         77,  77,  76,  75,  75,  74,  73,  73,  72,  71, 
         71,  70,  69,  68,  68,  67,  66,  66,  65,  64, 
         64,  63,  62,  61,  61,  60,  59,  58,  58,  57, 
         56,  55,  55,  54,  53,  52,  52,  51,  50,  49, 
         49,  48,  47,  46,  46,  45,  44,  43,  42,  42, 
         41,  40,  39,  38,  38,  37,  36,  35,  34,  34, 
         33,  32,  31,  30,  30,  29,  28,  27,  26,  26, 
         25,  24,  23,  22,  22,  21,  20,  19,  18,  17, 
         17,  16,  15,  14,  13,  12,  12,  11,  10,   9, 
          8,   7,   7,   6,   5,   4,   3,   2,   2,   1, 
          0 };
// The table constant cos_data contains a 241-point cosine table between 0
// and pi/4 (240). The values are integers between 0 and 127. This allows
// calculating a 960-point full cosine function that fits in 8 bits.
// A 960 point waveform sampled at 9600 Hz allows a 10 Hz frequency
// resolution (step size), which is enough for this application.
// The table is 241 points in size instead of 240 because the end point of
// each section is the same as the starting point of the following one.
// --------------------------------------------------------------------------




// --------------------------------------------------------------------------
// Global variables.
// --------------------------------------------------------------------------
unsigned int phase;                                             // This is the real time phase value that is recalculated at each cycle.
unsigned int delta_phi;                                         // This is the phase step that sets the frequency.


	

// --------------------------------------------------------------------------
// This function return an integer (8-bit) shifted cosine of the variable
// a (in the 0..959 range, which stands for 0..2*pi). The output is alwais
// positive and is in the 0..255 range. This allows direct driving of the
// output without converting negative numbers.
// The code is arranged in such a way that it will execute in "almost" the
// same time regardless of the value of a.
// --------------------------------------------------------------------------
unsigned char int_shifted_cos(unsigned int a)
{
    if (a <= 480)
    {
        if (a <= 240)
        {
            return (cos_data[(unsigned char)a] + 127);
        }
        else
        {
            return (-cos_data[(unsigned char)(480-a)] + 127);
        }        
    }
    else
    {
        if (a <= 720)
        {
            return (-cos_data[(unsigned char)(a-480)] + 127);
        }
        else
        {
            return (cos_data[(unsigned char)(960-a)] + 127);
        }
    }
}




// --------------------------------------------------------------------------
// This procedure is executed by the TMR0 interrupt that is triggered 9600 times
// per second (9.830400 MHz / (4 * 256)). This will be the output waveform
// sampling frequency. The output frequency can range from 0 to 4800 Hz (half the
// sampling frequency). The interrupt must be shorter than 256 instructions, since
// this is the time between two consecutive interrupts.
// --------------------------------------------------------------------------
void interrupt interrupt_handler(void)
{
    GIE = 0;                                                    // Disable interrupts.

    if (T0IF == 1)                                              // TMR0 overflow?
    {
        OUTPUT = int_shifted_cos(phase);                        // First output wave data (to avoid jitter because of unknow execution time of different "if" statements).

        phase += delta_phi;                                     // Than compute new phase for next interrupt.
        if (phase >= 960)                                       // Phase overflows at 960 (like 2*pi).
            phase -= 960;

        T0IF = 0;                                               // Clear TMR0 interrupt flag.
    }
    GIE = 1;                                                    // Enable interrupts.
}




// --------------------------------------------------------------------------
// This is the main program. 
// --------------------------------------------------------------------------
void main(void)
{
    PORTA = 0;              // Reset PORTA.
    TRISA = 0b00011111;     // PORTA: all input.

    OUTPUT = 127;           // Prepare output for mid range value.
    TRISB = 0b00000000;     // PORTB: all output.

    T0CS = 0;               // Drive TMR0 with FOSC/4
    PSA = 1;                // Do not use prescaler (TMR0 already divides by 256).
    T0IF = 0;               // Clear TMR0 interrupt flag.
    T0IE = 1;               // Enable TMR0 interrupt.

    phase = 0;              // Initialize phase.
    delta_phi = 0;          // Set phase increment to 0 (this will set the frequency to 0).

    GIE = 1;                // Enable interrupts to start the DDS.

    while (1)               // Endless loop (the frequency generation is interrupt driven).
    {
        switch (FREQ_SEL)   // Poll inputs to set the required frequency.
        {
            case 15: delta_phi = 30;  break;     // Dips-switch 0: 300 Hz
            case 14: delta_phi = 40;  break;     // Dips-switch 1: 400 Hz
            case 13: delta_phi = 44;  break;     // Dips-switch 2: 440 Hz
            case 12: delta_phi = 50;  break;     // Dips-switch 3: 500 Hz
            case 11: delta_phi = 60;  break;     // Dips-switch 4: 600 Hz
            case 10: delta_phi = 70;  break;     // Dips-switch 5: 700 Hz
            case 9:  delta_phi = 75;  break;     // Dips-switch 6: 750 Hz
            case 8:  delta_phi = 80;  break;     // Dips-switch 7: 800 Hz
            case 7:  delta_phi = 88;  break;     // Dips-switch 8: 880 Hz
            case 6:  delta_phi = 100; break;     // Dips-switch 9: 1000 Hz
            case 5:  delta_phi = 132; break;     // Dips-switch A: 1320 Hz
            case 4:  delta_phi = 175; break;     // Dips-switch B: 1750 Hz
            case 3:  delta_phi = 200; break;     // Dips-switch C: 2000 Hz
            case 2:  delta_phi = 250; break;     // Dips-switch D: 2500 Hz
            case 1:  delta_phi = 300; break;     // Dips-switch E: 3000 Hz
            case 0:  delta_phi = 330; break;     // Dips-switch F: 3300 Hz
            default: delta_phi = 0;
        }
    }
}
